local s = Skada
if s == nil then
	return
end

local AddonName, a = ...
local L = a.Localize

local COMBATLOG_OBJECT_AFFILIATION_OUTSIDER = COMBATLOG_OBJECT_AFFILIATION_OUTSIDER
local COMBATLOG_OBJECT_CONTROL_PLAYER = COMBATLOG_OBJECT_CONTROL_PLAYER
local COMBATLOG_OBJECT_REACTION_MASK = COMBATLOG_OBJECT_REACTION_MASK
local UnitBuff = UnitBuff
local UnitClass = UnitClass
local UnitGUID = UnitGUID
local UnitName = UnitName
local bit = bit
local math = math
local pairs = pairs
local select = select
local string = string

local function getTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		outerTable = outerTable[key]
		if outerTable == nil then
			return nil
		end
	end
	return outerTable
end

local function getOrMakeTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		if outerTable[key] == nil then
			outerTable[key] = {}
		end
		outerTable = outerTable[key]
	end
	return outerTable
end

local function updateBars(win, columnFlags, data, getAmount, getTime)
    local numBars = 0
    local total = 0
    
	win.metadata.wipestale = true
	win.metadata.maxvalue = 0
	
	-- set the values for each bar
	for _, info in pairs(data) do
		local amount = getAmount(info)
		if amount > 0 then
			numBars = numBars + 1
--if numBars < #win.dataset then
--print(win.dataset[numBars].label, "->", info.name)
--end
		    local bar = getOrMakeTable(win.dataset, numBars)
		    bar.id = info.id
		    bar.label = info.name
		    bar.value = amount
		    bar.activetime = getTime(info)
		    bar.class = info.class
		    
		    win.metadata.maxvalue = math.max(win.metadata.maxvalue, amount)
			total = total + amount
		end
	end
	
	while #win.dataset > numBars do
--print("trim")
		win.dataset[#win.dataset] = nil
	end
--	for i = numBars + 1, #win.dataset do
--print("zero", win.dataset[i].label)
--		win.dataset[i].value = 0
--	end
	
	-- set the text for each bar (in a separate loop so we know %s)
    for i = 1, numBars do
        local bar = win.dataset[i]
        bar.valuetext = s:FormatValueText(
            s:FormatNumber(bar.value),
            columnFlags.Healing,
            string.format(
                "%02.1f", bar.value / math.max(1, bar.activetime)),
            columnFlags.HPS,
            string.format(
                "%02.1f%%", bar.value / total * 100),
            columnFlags.Percent)
    end
end

--------------------------------------------------------- Healing Required Mode
local healingRequiredMode = s:NewModule(L["Healing Required"])
local sourceSubmenu = s:NewModule(L["Source Details"])
local spellSubmenu = s:NewModule(L["Spell Details"])

healingRequiredMode.metadata = {
    showspots = true,
    click1 = sourceSubmenu,
    columns = { Healing = true, HPS = true, Percent = true }
}

function healingRequiredMode:Update(win, set)
	updateBars(
		win, 
		self.metadata.columns, 
		set.players, 
		function(player)
			return player.healingRequired
		end,
		function(player)
			return s:PlayerActiveTime(set, player)
		end)
end

function healingRequiredMode:GetSetSummary(set)
	return s:FormatNumber(set.healingRequired)
end

---------------------------------------------------------------- Source Submenu
sourceSubmenu.metadata = {
    showspots = true,
    click1 = spellSubmenu,
    columns = { Healing = true, HPS = true, Percent = true }
}

function sourceSubmenu:Enter(win, id, label)
    self.targetID = id
    self.title = L["Healing done to"] .. " " .. label
end

function sourceSubmenu:Update(win, set)
	local player = s:find_player(set, self.targetID)
	updateBars(
		win, 
		self.metadata.columns, 
		player.healingRequiredSources, 
		function(source)
			return source.healing
		end,
		function(source)
			return s:PlayerActiveTime(set, player)
		end)
end

----------------------------------------------------------------- Spell Submenu
spellSubmenu.metadata = {
    showspots = true,
    columns = { Healing = true, HPS = true, Percent = true }
}

function spellSubmenu:Enter(win, id, label)
    self.sourceName = label
    self.title = string.format(
    	"%s %s %s %s", 
    	L["Healing done to"],
    	s:find_player(win:get_selected_set(), sourceSubmenu.targetID).name,
    	L["by"],
    	label)
end

function spellSubmenu:Update(win, set)
	local player = s:find_player(set, sourceSubmenu.targetID)
	updateBars(
		win, 
		self.metadata.columns, 
		player.healingRequiredSources[self.sourceName].spells, 
		function(spell)
			return spell.healing
		end,
		function(spell)
			return s:PlayerActiveTime(set, player)
		end)
end

-------------------------------------------------------------------- Combat Log
local shields = {}
local multipliers = {
	[55233] = 1, -- Vampiric Blood
	[105588] = 1, -- Vampiric Brood
	[22842] = 3, -- Frenzied Regeneration
	[105737] = 3, -- Mass Regeneration
}
local sourceInfo = {}
local targetInfo = {}
local bufferInfo = {}

local function record(set, spellID, spellName, amount, adjustedSourceInfo)
	
	local player = s:get_player(set, targetInfo.playerid, targetInfo.playername)
	if not player then
		return
	end
	
	local required = amount
	if adjustedSourceInfo.playerid == targetInfo.playerid then
		required = 0
	end
	
	-- total
	set.healingRequired = set.healingRequired + required
	
	-- player
	player.healingRequired = player.healingRequired + required
	
	-- source
	local source = player.healingRequiredSources[adjustedSourceInfo.playername]
	if source == nil then
		source = {
			id = adjustedSourceInfo.playerid, 
			name = adjustedSourceInfo.playername, 
			class = select(2, UnitClass(adjustedSourceInfo.playername)), 
			healing = 0, 
			spells = {} }
		player.healingRequiredSources[adjustedSourceInfo.playername] = source
	end
	source.healing = source.healing + amount
	
	-- spell
	local spell = source.spells[spellName]
	if spell == nil then
		spell = { id = spellID, name = spellName, healing = 0}
		source.spells[spellName] = spell
	end
	spell.healing = spell.healing + amount
end

local function maybeRecord(spellID, spellName, amount)
	if amount <= 0 then
		return
	end
	
	sourceInfo.spellname = spellName
	s:FixPets(sourceInfo)
	s:FixPets(targetInfo)
	spellName = sourceInfo.spellname
	
	for i = 1, 99999 do
		local buffName, _, _, _, _, _, _, caster, _, _, buffID, _, _, v1, v2, v3
			= UnitBuff(targetInfo.playername, i)
		if buffName == nil then
			break
		end
		
		local multiplier = 0
		if multipliers[buffID] == 1 then
			multiplier = v1
		elseif multipliers[buffID] == 3 then
			multiplier = v3
		end
		
		if multiplier > 0 then
			local buffAmount = amount / (1 + 100 / multiplier)
			bufferInfo.playerid = UnitGUID(caster)
			bufferInfo.playername = UnitName(caster)
			record(s.current, buffID, buffName, buffAmount, bufferInfo)
			record(s.total, buffID, buffName, buffAmount, bufferInfo)
			amount = amount - buffAmount
		end
	end
	
	record(s.current, spellID, spellName, amount, sourceInfo)
	record(s.total, spellID, spellName, amount, sourceInfo)
end

local function parseHeal(
	timestamp, 
	eventtype, 
	sourceID, 
	sourceName, 
	sourceFlags, 
	targetID, 
	targetName, 
	targetFlags, 
	spellID, 
	spellName, 
	spellSchool, 
	amount, 
	overhealing, 
	absorbed, 
	critical)
	
	sourceInfo.playerid = sourceID
	sourceInfo.playername = sourceName
	sourceInfo.playerflags = sourceFlags
	targetInfo.playerid = targetID
	targetInfo.playername = targetName
	targetInfo.playerflags = targetFlags
	maybeRecord(spellID, spellName, amount - overhealing + absorbed)
end

local function parseAuraApplied(
	timestamp, 
	eventtype, 
	sourceID, 
	sourceName, 
	sourceFlags, 
	targetID, 
	targetName, 
	targetFlags, 
	spellID, 
	spellName, 
	spellSchool, 
	auraType, 
	amount)

	if amount ~= nil
		and bit.band(sourceFlags, targetFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) ~= 0
		and bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER) == 0
		and bit.band(targetFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER) == 0
		and bit.band(sourceFlags, targetFlags, COMBATLOG_OBJECT_REACTION_MASK) ~= 0
		then
		
		getOrMakeTable(shields, sourceID, targetID)[spellID] = amount
	end
end

local function parseAuraRefreshed(
	timestamp, 
	eventtype, 
	sourceID, 
	sourceName, 
	sourceFlags, 
	targetID, 
	targetName, 
	targetFlags, 
	spellID, 
	spellName, 
	spellSchool, 
	auraType, 
	amount)
	
	if amount == nil then
		return
	end
	
	local t = getTable(shields, sourceID, targetID)
	if t == nil then
		return
	end
	
	if t[spellID] then
		sourceInfo.playerid = sourceID
		sourceInfo.playername = sourceName
		sourceInfo.playerflags = sourceFlags
		targetInfo.playerid = targetID
		targetInfo.playername = targetName
		targetInfo.playerflags = targetFlags
		maybeRecord(spellID, spellName, t[spellID] - amount)
	end
	t[spellID] = amount
end

local function parseAuraRemoved(
	timestamp, 
	eventtype, 
	sourceID, 
	sourceName, 
	sourceFlags, 
	targetID, 
	targetName, 
	targetFlags, 
	spellID, 
	spellName, 
	spellSchool, 
	auraType, 
	amount)
	
	parseAuraRefreshed(
		timestamp, 
		eventtype, 
		sourceID, 
		sourceName, 
		sourceFlags, 
		targetID, 
		targetName, 
		targetFlags, 
		spellID, 
		spellName, 
		spellSchool, 
		auraType, 
		amount)
	
	local t = getTable(shields, sourceID, targetID)
	if t ~= nil then
		t[spellID] = nil
	end
end

-------------------------------------------------------------------------- Init
function healingRequiredMode:AddSetAttributes(set)
	set.healingRequired = set.healingRequired or 0
end

function healingRequiredMode:AddPlayerAttributes(player)
	player.healingRequired = player.healingRequired or 0
	getOrMakeTable(player, "healingRequiredSources")
	for _, source in pairs(player.healingRequiredSources) do
		source.healing = source.healing or 0
		for _, spell in pairs(getOrMakeTable(source, "spells")) do
			spell.healing = spell.healing or 0
		end
	end
end

function healingRequiredMode:OnEnable()
	s:RegisterForCL(parseHeal, 'SPELL_HEAL', {src_is_interesting = true})
	s:RegisterForCL(
		parseHeal, 'SPELL_PERIODIC_HEAL', {src_is_interesting = true})
	s:RegisterForCL(
		parseAuraApplied, 'SPELL_AURA_APPLIED', {src_is_interesting = true})
	s:RegisterForCL(
		parseAuraRefreshed, 'SPELL_AURA_REFRESH', {src_is_interesting = true})
	s:RegisterForCL(
		parseAuraRemoved, 'SPELL_AURA_REMOVED', {src_is_interesting = true})
	s:AddMode(healingRequiredMode)
end

function healingRequiredMode:OnDisable()
	s:RemoveMode(healingRequiredMode)
end
